@@ -23,15 +23,23 @@ import distutils.command.build
 import distutils.command.build_py
 import distutils.command.install
 import distutils.command.install_data
-import distutils.dist
 import distutils.dir_util
+import distutils.dist
 import glob
 import os.path
 import platform
 import sys
-
 from distutils.log import info
 
+try:
+    import distro
+except ImportError:
+    python_version = tuple(int(x) for x in platform.python_version_tuple())
+    if python_version >= (3, 8):
+        print(
+            'Missing build requirement "distro" Python module; '
+            'install paths may be incorrect', file=sys.stderr)
+
 
 def has_help(self):
     return "build_help" in self.distribution.cmdclass and os.name != 'nt'
@@ -66,7 +74,7 @@ class MeldDistribution(distutils.dist.Distribution):
     def __init__(self, *args, **kwargs):
         self.no_update_icon_cache = False
         self.no_compile_schemas = False
-        distutils.dist.Distribution.__init__(self, *args, **kwargs)
+        super().__init__(*args, **kwargs)
 
 
 class build_data(distutils.cmd.Command):
@@ -228,7 +236,7 @@ class build_i18n(distutils.cmd.Command):
     # way except magically extracting them from self.distribution.data_files
     desktop_files = [('share/applications', glob.glob("data/*.desktop.in"))]
     xml_files = [
-        ('share/appdata', glob.glob("data/*.appdata.xml.in")),
+        ('share/metainfo', glob.glob("data/*.appdata.xml.in")),
         ('share/mime/packages', glob.glob("data/mime/*.xml.in"))
     ]
     schemas_files = []
@@ -259,7 +267,10 @@ class build_i18n(distutils.cmd.Command):
 
         # Update po(t) files and print a report
         # We have to change the working dir to the po dir for intltool
-        cmd = ["intltool-update", (self.merge_po and "-r" or "-p"), "-g", self.domain]
+        cmd = [
+            "intltool-update",
+            (self.merge_po and "-r" or "-p"), "-g", self.domain
+        ]
         wd = os.getcwd()
         os.chdir(self.po_dir)
         self.spawn(cmd)
@@ -275,7 +286,8 @@ class build_i18n(distutils.cmd.Command):
                 os.makedirs(mo_dir)
             cmd = [msgfmt, po_file, "-o", mo_file]
             po_mtime = os.path.getmtime(po_file)
-            mo_mtime = os.path.exists(mo_file) and os.path.getmtime(mo_file) or 0
+            mo_mtime = (
+                os.path.exists(mo_file) and os.path.getmtime(mo_file) or 0)
             if po_mtime > max_po_mtime:
                 max_po_mtime = po_mtime
             if po_mtime > mo_mtime:
@@ -315,7 +327,8 @@ class build_i18n(distutils.cmd.Command):
                     mtime_merged = (os.path.exists(file_merged) and
                                     os.path.getmtime(file_merged) or 0)
                     mtime_file = os.path.getmtime(file)
-                    if mtime_merged < self.max_po_mtime or mtime_merged < mtime_file:
+                    if (mtime_merged < self.max_po_mtime or
+                            mtime_merged < mtime_file):
                         # Only build if output is older than input (.po,.in)
                         self.spawn(cmd)
                     files_merged.append(file_merged)
@@ -341,7 +354,7 @@ class build_py(distutils.command.build_py.build_py):
                 options = self.distribution.get_option_dict('install')
                 prefix = options['prefix'][1]
             except KeyError as e:
-                print (e)
+                print(e)
                 prefix = sys.prefix
 
             datadir = os.path.join(prefix, 'share', 'meld')
@@ -374,11 +387,21 @@ class install(distutils.command.install.install):
 
     def finalize_options(self):
         special_cases = ('debian', 'ubuntu', 'linuxmint')
-        if (platform.system() == 'Linux' and
-                platform.linux_distribution()[0].lower() in special_cases):
-            # Maintain an explicit install-layout, but use deb by default
-            specified_layout = getattr(self, 'install_layout', None)
-            self.install_layout = specified_layout or 'deb'
+        if platform.system() == 'Linux':
+            # linux_distribution has been removed in Python 3.8; we require
+            # the third-party distro package for future handling.
+            try:
+                distribution = platform.linux_distribution()[0].lower()
+            except AttributeError:
+                try:
+                    distribution = distro.id()
+                except NameError:
+                    distribution = 'unknown'
+
+            if distribution in special_cases:
+                # Maintain an explicit install-layout, but use deb by default
+                specified_layout = getattr(self, 'install_layout', None)
+                self.install_layout = specified_layout or 'deb'
 
         distutils.command.install.install.finalize_options(self)
 
