@@ -19,16 +19,17 @@ import logging
 import optparse
 import os
 
+from gi.repository import Gdk
 from gi.repository import Gio
 from gi.repository import GLib
-from gi.repository import Gdk
 from gi.repository import Gtk
 
 import meld.conf
 import meld.preferences
 import meld.ui.util
-
 from meld.conf import _
+from meld.filediff import FileDiff
+from meld.meldwindow import MeldWindow
 
 log = logging.getLogger(__name__)
 
@@ -42,10 +43,12 @@ optparse._ = _
 class MeldApp(Gtk.Application):
 
     def __init__(self):
-        Gtk.Application.__init__(self)
-        self.set_flags(Gio.ApplicationFlags.HANDLES_COMMAND_LINE)
-        self.set_application_id("org.gnome.meld")
+        super().__init__(
+          application_id=meld.conf.APPLICATION_ID,
+          flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
+        )
         GLib.set_application_name("Meld")
+        GLib.set_prgname(meld.conf.APPLICATION_ID)
         Gtk.Window.set_default_icon_name("meld")
 
     def do_startup(self):
@@ -78,6 +81,7 @@ class MeldApp(Gtk.Application):
         if isinstance(tab, int):
             return tab
         elif tab:
+
             def done(tab, status):
                 self.release()
                 tab.command_line.set_exit_status(status)
@@ -109,12 +113,12 @@ class MeldApp(Gtk.Application):
         meld.preferences.PreferencesDialog(self.get_active_window())
 
     def help_callback(self, action, parameter):
-        if meld.conf.UNINSTALLED:
+        if meld.conf.DATADIR_IS_UNINSTALLED:
             uri = "http://meldmerge.org/help/"
         else:
             uri = "help:meld"
-        Gtk.show_uri(Gdk.Screen.get_default(), uri,
-                     Gtk.get_current_event_time())
+        Gtk.show_uri(
+            Gdk.Screen.get_default(), uri, Gtk.get_current_event_time())
 
     def about_callback(self, action, parameter):
         about = meld.ui.util.get_widget("application.ui", "aboutdialog")
@@ -125,15 +129,15 @@ class MeldApp(Gtk.Application):
 
     def quit_callback(self, action, parameter):
         for window in self.get_windows():
-            cancelled = window.emit("delete-event",
-                                    Gdk.Event.new(Gdk.EventType.DELETE))
+            cancelled = window.emit(
+                "delete-event", Gdk.Event.new(Gdk.EventType.DELETE))
             if cancelled:
                 return
             window.destroy()
         self.quit()
 
     def new_window(self):
-        window = meldwindow.MeldWindow()
+        window = MeldWindow()
         self.add_window(window.widget)
         window.widget.meldwindow = window
         return window
@@ -141,18 +145,20 @@ class MeldApp(Gtk.Application):
     def get_meld_window(self):
         return self.get_active_window().meldwindow
 
-    def open_files(self, files, **kwargs):
-        new_tab = kwargs.pop('new_tab')
-        if new_tab:
-            window = self.get_meld_window()
-        else:
-            window = self.new_window()
+    def open_files(
+            self, gfiles, *, window=None, close_on_error=False, **kwargs):
+        """Open a comparison between files in a Meld window
 
-        paths = [f.get_path() for f in files]
+        :param gfiles: list of Gio.File to be compared
+        :param window: window in which to open comparison tabs; if
+            None, the current window is used
+        :param close_on_error: if true, close window if an error occurs
+        """
+        window = window or self.get_meld_window()
         try:
-            return window.open_paths(paths, **kwargs)
+            return window.open_paths(gfiles, **kwargs)
         except ValueError:
-            if not new_tab:
+            if close_on_error:
                 self.remove_window(window.widget)
             raise
 
@@ -196,7 +202,7 @@ class MeldApp(Gtk.Application):
                 self.should_exit = False
                 self.output = io.StringIO()
                 self.exit_status = 0
-                optparse.OptionParser.__init__(self, *args, **kwargs)
+                super().__init__(*args, **kwargs)
 
             def exit(self, status=0, msg=None):
                 self.should_exit = True
@@ -205,7 +211,7 @@ class MeldApp(Gtk.Application):
                 try:
                     self.command_line.do_print_literal(
                         self.command_line, self.output.getvalue())
-                except:
+                except Exception:
                     print(self.output.getvalue())
                 self.exit_status = status
 
@@ -284,9 +290,9 @@ class MeldApp(Gtk.Application):
             parser.local_error(_("too many arguments (wanted 0-3, got %d)") %
                                len(args))
         elif options.auto_merge and len(args) < 3:
-            parser.local_error(_("can't auto-merge less than 3 files"))
+            parser.local_error(_("can’t auto-merge less than 3 files"))
         elif options.auto_merge and any([os.path.isdir(f) for f in args]):
-            parser.local_error(_("can't auto-merge directories"))
+            parser.local_error(_("can’t auto-merge directories"))
 
         if parser.should_exit:
             cleanup()
@@ -315,23 +321,49 @@ class MeldApp(Gtk.Application):
                 if relative.query_exists(cancellable=None):
                     return relative
                 # Return the original arg for a better error message
+
+            if f.get_uri() is None:
+                raise ValueError(_("invalid path or URI “%s”") % arg)
+
+            # TODO: support for directories specified by URIs
+            file_type = f.query_file_type(Gio.FileQueryInfoFlags.NONE, None)
+            if not f.is_native() and file_type == Gio.FileType.DIRECTORY:
+                raise ValueError(
+                    _("remote folder “{}” not supported").format(arg))
+
             return f
 
         tab = None
         error = None
-        comparisons = [args] + options.diff
-        options.newtab = options.newtab or not command_line.get_is_remote()
+        comparisons = [c for c in [args] + options.diff if c]
+
+        # Every Meld invocation creates at most one window. If there is
+        # no existing application, a window is created in do_startup().
+        # If there is an existing application, then this is a remote
+        # invocation, in which case we'll create a window if and only
+        # if the new-tab flag is not provided.
+        #
+        # In all cases, all tabs newly created here are attached to the
+        # same window, either implicitly by using the most-recently-
+        # focused window, or explicitly as below.
+        window = None
+        close_on_error = False
+        if command_line.get_is_remote() and not options.newtab:
+            window = self.new_window()
+            close_on_error = True
+
         for i, paths in enumerate(comparisons):
-            files = [make_file_from_command_line(p) for p in paths]
             auto_merge = options.auto_merge and i == 0
             try:
-                for p, f in zip(paths, files):
-                    if f.get_path() is None:
-                        raise ValueError(_("invalid path or URI \"%s\"") % p)
+                files = [make_file_from_command_line(p) for p in paths]
                 tab = self.open_files(
-                    files, auto_compare=options.auto_compare,
-                    auto_merge=auto_merge, new_tab=options.newtab,
-                    focus=i == 0)
+                    files,
+                    window=window,
+                    close_on_error=close_on_error,
+                    auto_compare=options.auto_compare,
+                    auto_merge=auto_merge,
+                    focus=i == 0,
+                )
             except ValueError as err:
                 error = err
                 log.debug("Couldn't open comparison: %s", error, exc_info=True)
@@ -342,9 +374,9 @@ class MeldApp(Gtk.Application):
                 if options.label:
                     tab.set_labels(options.label)
 
-                if options.outfile and isinstance(tab, filediff.FileDiff):
+                if options.outfile and isinstance(tab, FileDiff):
                     outfile = make_file_from_command_line(options.outfile)
-                    tab.set_merge_output_file(outfile.get_path())
+                    tab.set_merge_output_file(outfile)
 
         if error:
             if not tab:
@@ -364,6 +396,3 @@ class MeldApp(Gtk.Application):
 
 
 app = MeldApp()
-
-from . import filediff
-from . import meldwindow
