@@ -22,38 +22,36 @@ from gi.repository import GLib
 from gi.repository import Gtk
 
 import meld.ui.util
-from . import dirdiff
-from . import filediff
-from . import filemerge
-from . import melddoc
-from . import newdifftab
-from . import recent
-from . import task
-from . import vcview
-from .ui import gnomeglade
-from .ui import notebooklabel
-
 from meld.conf import _
-from meld.recent import recent_comparisons
+from meld.dirdiff import DirDiff
+from meld.filediff import FileDiff
+from meld.filemerge import FileMerge
+from meld.melddoc import ComparisonState, MeldDoc
+from meld.newdifftab import NewDiffTab
+from meld.recent import recent_comparisons, RecentType
 from meld.settings import interface_settings, settings
+from meld.task import LifoScheduler
+from meld.ui.gnomeglade import Component, ui_file
+from meld.ui.notebooklabel import NotebookLabel
+from meld.vcview import VcView
 from meld.windowstate import SavedWindowState
 
 
-class MeldWindow(gnomeglade.Component):
+class MeldWindow(Component):
 
     def __init__(self):
-        gnomeglade.Component.__init__(self, "meldapp.ui", "meldapp")
+        super().__init__("meldapp.ui", "meldapp")
         self.widget.set_name("meldapp")
 
         actions = (
             ("FileMenu", None, _("_File")),
-            ("New", Gtk.STOCK_NEW, _("_New Comparison..."), "<Primary>N",
+            ("New", Gtk.STOCK_NEW, _("_New Comparison…"), "<Primary>N",
                 _("Start a new comparison"),
                 self.on_menu_file_new_activate),
             ("Save", Gtk.STOCK_SAVE, None, None,
                 _("Save the current file"),
                 self.on_menu_save_activate),
-            ("SaveAs", Gtk.STOCK_SAVE_AS, _("Save As..."), "<Primary><shift>S",
+            ("SaveAs", Gtk.STOCK_SAVE_AS, _("Save As…"), "<Primary><shift>S",
                 _("Save the current file with a different name"),
                 self.on_menu_save_as_activate),
             ("Close", Gtk.STOCK_CLOSE, None, None,
@@ -73,7 +71,7 @@ class MeldWindow(gnomeglade.Component):
                 self.on_menu_copy_activate),
             ("Paste", Gtk.STOCK_PASTE, None, None, _("Paste the clipboard"),
                 self.on_menu_paste_activate),
-            ("Find", Gtk.STOCK_FIND, _("Find..."), None, _("Search for text"),
+            ("Find", Gtk.STOCK_FIND, _("Find…"), None, _("Search for text"),
                 self.on_menu_find_activate),
             ("FindNext", None, _("Find Ne_xt"), "<Primary>G",
                 _("Search forwards for the same text"),
@@ -82,9 +80,12 @@ class MeldWindow(gnomeglade.Component):
                 _("Search backwards for the same text"),
                 self.on_menu_find_previous_activate),
             ("Replace", Gtk.STOCK_FIND_AND_REPLACE,
-                _("_Replace..."), "<Primary>H",
+                _("_Replace…"), "<Primary>H",
                 _("Find and replace text"),
                 self.on_menu_replace_activate),
+            ("GoToLine", None, _("Go to _Line"), "<Primary>I",
+                _("Go to a specific line"),
+                self.on_menu_go_to_line_activate),
 
             ("ChangesMenu", None, _("_Changes")),
             ("NextChange", Gtk.STOCK_GO_DOWN, _("Next Change"), "<Alt>Down",
@@ -117,7 +118,6 @@ class MeldWindow(gnomeglade.Component):
                 _("Show or hide the toolbar"),
                 None, True),
         )
-        ui_file = gnomeglade.ui_file("meldapp-ui.xml")
         self.actiongroup = Gtk.ActionGroup(name='MainActions')
         self.actiongroup.set_translation_domain("meld")
         self.actiongroup.add_actions(actions)
@@ -134,7 +134,7 @@ class MeldWindow(gnomeglade.Component):
 
         self.ui = Gtk.UIManager()
         self.ui.insert_action_group(self.actiongroup, 0)
-        self.ui.add_ui_from_file(ui_file)
+        self.ui.add_ui_from_file(ui_file("meldapp-ui.xml"))
 
         # Manually handle shells that don't show an application menu
         gtk_settings = Gtk.Settings.get_default()
@@ -164,8 +164,7 @@ class MeldWindow(gnomeglade.Component):
             app_actiongroup.add_actions(app_actions)
             self.ui.insert_action_group(app_actiongroup, 0)
 
-            ui_file = gnomeglade.ui_file("appmenu-fallback.xml")
-            self.ui.add_ui_from_file(ui_file)
+            self.ui.add_ui_from_file(ui_file("appmenu-fallback.xml"))
             self.widget.set_show_menubar(False)
 
         for menuitem in ("Save", "Undo"):
@@ -184,14 +183,19 @@ class MeldWindow(gnomeglade.Component):
         interface_settings.bind('toolbar-style', self.toolbar, 'toolbar-style',
                                 Gio.SettingsBindFlags.DEFAULT)
 
-        # Add alternate keybindings for Prev/Next Change
-        accels = self.ui.get_accel_group()
-        (keyval, mask) = Gtk.accelerator_parse("<Primary>D")
-        accels.connect(keyval, mask, 0, self.on_menu_edit_down_activate)
-        (keyval, mask) = Gtk.accelerator_parse("<Primary>E")
-        accels.connect(keyval, mask, 0, self.on_menu_edit_up_activate)
-        (keyval, mask) = Gtk.accelerator_parse("F5")
-        accels.connect(keyval, mask, 0, self.on_menu_refresh_activate)
+        # Alternate keybindings for a few commands.
+        extra_accels = (
+            ("<Primary>D", self.on_menu_edit_down_activate),
+            ("<Primary>E", self.on_menu_edit_up_activate),
+            ("<Alt>KP_Down", self.on_menu_edit_down_activate),
+            ("<Alt>KP_Up", self.on_menu_edit_up_activate),
+            ("F5", self.on_menu_refresh_activate),
+        )
+
+        accel_group = self.ui.get_accel_group()
+        for accel, callback in extra_accels:
+            keyval, mask = Gtk.accelerator_parse(accel)
+            accel_group.connect(keyval, mask, 0, callback)
 
         # Initialise sensitivity for important actions
         self.actiongroup.get_action("Stop").set_sensitive(False)
@@ -215,8 +219,13 @@ class MeldWindow(gnomeglade.Component):
             action.connect('activate', callback)
             self.widget.add_action(action)
 
+        # Create a secondary toolbar, just to hold our progress spinner
         toolbutton = Gtk.ToolItem()
         self.spinner = Gtk.Spinner()
+        # Fake out the spinner on Windows. See Gitlab issue #133.
+        if os.name == 'nt':
+            for attr in ('stop', 'hide', 'show', 'start'):
+                setattr(self.spinner, attr, lambda *args: True)
         toolbutton.add(self.spinner)
         self.secondary_toolbar.insert(toolbutton, -1)
         # Set a minimum size because the spinner requests nothing
@@ -236,7 +245,7 @@ class MeldWindow(gnomeglade.Component):
 
         self.should_close = False
         self.idle_hooked = 0
-        self.scheduler = task.LifoScheduler()
+        self.scheduler = LifoScheduler()
         self.scheduler.connect("runnable", self.on_scheduler_runnable)
 
         self.ui.ensure_update()
@@ -249,13 +258,9 @@ class MeldWindow(gnomeglade.Component):
         rmenu = self.ui.get_widget('/Menubar/FileMenu/Recent').get_submenu()
         rmenu.connect("map", self._on_recentmenu_map)
 
-        try:
-            builder = meld.ui.util.get_builder("shortcuts.ui")
-            shortcut_window = builder.get_object("shortcuts-meld")
-            self.widget.set_help_overlay(shortcut_window)
-        except GLib.Error:
-            # GtkShortcutsWindow is new in GTK+ 3.20
-            pass
+        builder = meld.ui.util.get_builder("shortcuts.ui")
+        shortcut_window = builder.get_object("shortcuts-meld")
+        self.widget.set_help_overlay(shortcut_window)
 
     def _on_recentmenu_map(self, recentmenu):
         for imagemenuitem in recentmenu.get_children():
@@ -269,13 +274,11 @@ class MeldWindow(gnomeglade.Component):
         # Let the rest of the stack know about this event
         return False
 
-    def on_widget_drag_data_received(self, wid, context, x, y, selection_data,
-                                     info, time):
-        if len(selection_data.get_uris()) != 0:
-            paths = []
-            for uri in selection_data.get_uris():
-                paths.append(Gio.File.new_for_uri(uri).get_path())
-            self.open_paths(paths)
+    def on_widget_drag_data_received(
+            self, wid, context, x, y, selection_data, info, time):
+        uris = selection_data.get_uris()
+        if uris:
+            self.open_paths([Gio.File.new_for_uri(uri) for uri in uris])
             return True
 
     def on_idle(self):
@@ -327,17 +330,17 @@ class MeldWindow(gnomeglade.Component):
             page = None
 
         self.actiongroup.get_action("Close").set_sensitive(bool(page))
-        if not isinstance(page, melddoc.MeldDoc):
+        if not isinstance(page, MeldDoc):
             for action in ("PrevChange", "NextChange", "Cut", "Copy", "Paste",
                            "Find", "FindNext", "FindPrevious", "Replace",
-                           "Refresh"):
+                           "Refresh", "GoToLine"):
                 self.actiongroup.get_action(action).set_sensitive(False)
         else:
             for action in ("Find", "Refresh"):
                 self.actiongroup.get_action(action).set_sensitive(True)
-            is_filediff = isinstance(page, filediff.FileDiff)
+            is_filediff = isinstance(page, FileDiff)
             for action in ("Cut", "Copy", "Paste", "FindNext", "FindPrevious",
-                           "Replace"):
+                           "Replace", "GoToLine"):
                 self.actiongroup.get_action(action).set_sensitive(is_filediff)
 
     def handle_current_doc_switch(self, page):
@@ -370,7 +373,7 @@ class MeldWindow(gnomeglade.Component):
         self.actiongroup.get_action("Redo").set_sensitive(can_redo)
 
         # FileDiff handles save sensitivity; it makes no sense for other modes
-        if not isinstance(newdoc, filediff.FileDiff):
+        if not isinstance(newdoc, FileDiff):
             self.actiongroup.get_action("Save").set_sensitive(False)
             self.actiongroup.get_action("SaveAs").set_sensitive(False)
         else:
@@ -383,7 +386,7 @@ class MeldWindow(gnomeglade.Component):
         else:
             self.widget.set_title("Meld")
 
-        if isinstance(newdoc, melddoc.MeldDoc):
+        if isinstance(newdoc, MeldDoc):
             self.diff_handler = newdoc.connect("next-diff-changed",
                                                self.on_next_diff_changed)
         else:
@@ -456,6 +459,9 @@ class MeldWindow(gnomeglade.Component):
     def on_menu_replace_activate(self, *extra):
         self.current_doc().on_replace_activate()
 
+    def on_menu_go_to_line_activate(self, *extra):
+        self.current_doc().on_go_to_line_activate()
+
     def on_menu_copy_activate(self, *extra):
         widget = self.widget.get_focus()
         if isinstance(widget, Gtk.Editable):
@@ -522,7 +528,7 @@ class MeldWindow(gnomeglade.Component):
                     self.widget.emit('destroy')
 
     def on_page_state_changed(self, page, old_state, new_state):
-        if self.should_close and old_state == melddoc.STATE_CLOSING:
+        if self.should_close and old_state == ComparisonState.Closing:
             # Cancel closing if one of our tabs does
             self.should_close = False
 
@@ -533,21 +539,20 @@ class MeldWindow(gnomeglade.Component):
                 page.on_file_changed(filename)
 
     def _append_page(self, page, icon):
-        nbl = notebooklabel.NotebookLabel(
-            icon, "", lambda b: page.on_delete_event())
+        nbl = NotebookLabel(icon, "", lambda b: page.on_delete_event())
         self.notebook.append_page(page.widget, nbl)
 
         # Change focus to the newly created page only if the user is on a
         # DirDiff or VcView page, or if it's a new tab page. This prevents
         # cycling through X pages when X diffs are initiated.
-        if isinstance(self.current_doc(), dirdiff.DirDiff) or \
-           isinstance(self.current_doc(), vcview.VcView) or \
-           isinstance(page, newdifftab.NewDiffTab):
+        if isinstance(self.current_doc(), DirDiff) or \
+           isinstance(self.current_doc(), VcView) or \
+           isinstance(page, NewDiffTab):
             self.notebook.set_current_page(self.notebook.page_num(page.widget))
 
         if hasattr(page, 'scheduler'):
             self.scheduler.add_scheduler(page.scheduler)
-        if isinstance(page, melddoc.MeldDoc):
+        if isinstance(page, MeldDoc):
             page.connect("file-changed", self.on_file_changed)
             page.connect("create-diff", lambda obj, arg, kwargs:
                          self.append_diff(arg, **kwargs))
@@ -557,7 +562,7 @@ class MeldWindow(gnomeglade.Component):
         self.notebook.set_tab_reorderable(page.widget, True)
 
     def append_new_comparison(self):
-        doc = newdifftab.NewDiffTab(self)
+        doc = NewDiffTab(self)
         self._append_page(doc, "document-new")
         self.notebook.on_label_changed(doc, _("New comparison"), None)
 
@@ -569,103 +574,111 @@ class MeldWindow(gnomeglade.Component):
         doc.connect("diff-created", diff_created_cb)
         return doc
 
-    def append_dirdiff(self, dirs, auto_compare=False):
+    def append_dirdiff(self, gfiles, auto_compare=False):
+        dirs = [d.get_path() if d else None for d in gfiles]
         assert len(dirs) in (1, 2, 3)
-        doc = dirdiff.DirDiff(len(dirs))
+        doc = DirDiff(len(dirs))
         self._append_page(doc, "folder")
         doc.set_locations(dirs)
-        # FIXME: This doesn't work, as dirdiff behaves differently to vcview
         if auto_compare:
-            doc.on_button_diff_clicked(None)
+            doc.scheduler.add_task(doc.auto_compare)
         return doc
 
-    def append_filediff(self, files, merge_output=None, meta=None):
-        assert len(files) in (1, 2, 3)
-        doc = filediff.FileDiff(len(files))
+    def append_filediff(
+            self, gfiles, *, encodings=None, merge_output=None, meta=None):
+        assert len(gfiles) in (1, 2, 3)
+        doc = FileDiff(len(gfiles))
         self._append_page(doc, "text-x-generic")
-        doc.set_files(files)
+        doc.set_files(gfiles, encodings)
         if merge_output is not None:
             doc.set_merge_output_file(merge_output)
         if meta is not None:
             doc.set_meta(meta)
         return doc
 
-    def append_filemerge(self, files, merge_output=None):
-        if len(files) != 3:
+    def append_filemerge(self, gfiles, merge_output=None):
+        if len(gfiles) != 3:
             raise ValueError(
-                _("Need three files to auto-merge, got: %r") % files)
-        doc = filemerge.FileMerge(len(files))
+                _("Need three files to auto-merge, got: %r") %
+                [f.get_parse_name() for f in gfiles])
+        doc = FileMerge(len(gfiles))
         self._append_page(doc, "text-x-generic")
-        doc.set_files(files)
+        doc.set_files(gfiles)
         if merge_output is not None:
             doc.set_merge_output_file(merge_output)
         return doc
 
-    def append_diff(self, paths, auto_compare=False, auto_merge=False,
+    def append_diff(self, gfiles, auto_compare=False, auto_merge=False,
                     merge_output=None, meta=None):
-        dirslist = [p for p in paths if os.path.isdir(p)]
-        fileslist = [p for p in paths if os.path.isfile(p)]
-        if dirslist and fileslist:
+        have_directories = False
+        have_files = False
+        for f in gfiles:
+            if f.query_file_type(
+               Gio.FileQueryInfoFlags.NONE, None) == Gio.FileType.DIRECTORY:
+                have_directories = True
+            else:
+                have_files = True
+        if have_directories and have_files:
             raise ValueError(
                 _("Cannot compare a mixture of files and directories"))
-        elif dirslist:
-            return self.append_dirdiff(paths, auto_compare)
+        elif have_directories:
+            return self.append_dirdiff(gfiles, auto_compare)
         elif auto_merge:
-            return self.append_filemerge(paths, merge_output=merge_output)
+            return self.append_filemerge(gfiles, merge_output=merge_output)
         else:
             return self.append_filediff(
-                paths, merge_output=merge_output, meta=meta)
+                gfiles, merge_output=merge_output, meta=meta)
 
     def append_vcview(self, location, auto_compare=False):
-        doc = vcview.VcView()
+        doc = VcView()
         self._append_page(doc, "meld-version-control")
-        location = location[0] if isinstance(location, list) else location
-        doc.set_location(location)
+        if isinstance(location, (list, tuple)):
+            location = location[0]
+        doc.set_location(location.get_path())
         if auto_compare:
-            doc.on_button_diff_clicked(None)
+            doc.scheduler.add_task(doc.auto_compare)
         return doc
 
     def append_recent(self, uri):
-        comparison_type, files, flags = recent_comparisons.read(uri)
-        if comparison_type == recent.TYPE_MERGE:
-            tab = self.append_filemerge(files)
-        elif comparison_type == recent.TYPE_FOLDER:
-            tab = self.append_dirdiff(files)
-        elif comparison_type == recent.TYPE_VC:
-            # Files should be a single-element iterable
-            tab = self.append_vcview(files[0])
-        else:  # comparison_type == recent.TYPE_FILE:
-            tab = self.append_filediff(files)
+        comparison_type, gfiles, flags = recent_comparisons.read(uri)
+        comparison_method = {
+            RecentType.File: self.append_filediff,
+            RecentType.Folder: self.append_dirdiff,
+            RecentType.Merge: self.append_filemerge,
+            RecentType.VersionControl: self.append_vcview,
+        }
+        tab = comparison_method[comparison_type](gfiles)
         self.notebook.set_current_page(self.notebook.page_num(tab.widget))
         recent_comparisons.add(tab)
         return tab
 
-    def _single_file_open(self, path):
-        doc = vcview.VcView()
+    def _single_file_open(self, gfile):
+        doc = VcView()
 
         def cleanup():
             self.scheduler.remove_scheduler(doc.scheduler)
         self.scheduler.add_task(cleanup)
         self.scheduler.add_scheduler(doc.scheduler)
-        path = os.path.abspath(path)
+        path = gfile.get_path()
         doc.set_location(path)
         doc.connect("create-diff", lambda obj, arg, kwargs:
                     self.append_diff(arg, **kwargs))
         doc.run_diff(path)
 
-    def open_paths(self, paths, auto_compare=False, auto_merge=False,
+    def open_paths(self, gfiles, auto_compare=False, auto_merge=False,
                    focus=False):
         tab = None
-        if len(paths) == 1:
-            a = paths[0]
-            if os.path.isfile(a):
-                self._single_file_open(a)
-            else:
+        if len(gfiles) == 1:
+            a = gfiles[0]
+            if a.query_file_type(Gio.FileQueryInfoFlags.NONE, None) == \
+                    Gio.FileType.DIRECTORY:
                 tab = self.append_vcview(a, auto_compare)
+            else:
+                self._single_file_open(a)
 
-        elif len(paths) in (2, 3):
-            tab = self.append_diff(
-                paths, auto_compare=auto_compare, auto_merge=auto_merge)
+        elif len(gfiles) in (2, 3):
+            tab = self.append_diff(gfiles, auto_compare=auto_compare,
+                                   auto_merge=auto_merge)
         if tab:
             recent_comparisons.add(tab)
             if focus:
@@ -679,10 +692,10 @@ class MeldWindow(gnomeglade.Component):
         index = self.notebook.get_current_page()
         if index >= 0:
             page = self.notebook.get_nth_page(index).pyobject
-            if isinstance(page, melddoc.MeldDoc):
+            if isinstance(page, MeldDoc):
                 return page
 
-        class DummyDoc(object):
+        class DummyDoc:
             def __getattr__(self, a):
                 return lambda *x: None
         return DummyDoc()
