@@ -15,8 +15,8 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import cairo
-from gi.repository import GObject
 from gi.repository import Gdk
+from gi.repository import GObject
 from gi.repository import Gtk
 
 
@@ -27,23 +27,27 @@ class EmblemCellRenderer(Gtk.CellRenderer):
     __gproperties__ = {
         "icon-name":   (str, "Named icon",
                         "Name for base icon",
-                        "text-x-generic", GObject.PARAM_READWRITE),
+                        "text-x-generic", GObject.ParamFlags.READWRITE),
         "emblem-name": (str, "Named emblem icon",
                         "Name for emblem icon to overlay",
-                        None, GObject.PARAM_READWRITE),
-        "icon-tint":   (str, "Icon tint",
+                        None, GObject.ParamFlags.READWRITE),
+        "secondary-emblem-name": (
+            str, "Named secondary emblem icon",
+            "Name for secondary emblem icon to overlay",
+            None, GObject.ParamFlags.READWRITE),
+        "icon-tint":   (Gdk.RGBA, "Icon tint",
                         "GDK-parseable color to be used to tint icon",
-                        None, GObject.PARAM_READWRITE),
+                        GObject.ParamFlags.READWRITE),
     }
 
     icon_cache = {}
 
     def __init__(self):
-        super(EmblemCellRenderer, self).__init__()
+        super().__init__()
         self._icon_name = "text-x-generic"
         self._emblem_name = None
+        self._secondary_emblem_name = None
         self._icon_tint = None
-        self._tint_color = None
         self._state = None
         # FIXME: hardcoded sizes
         self._icon_size = 16
@@ -54,13 +58,10 @@ class EmblemCellRenderer(Gtk.CellRenderer):
             self._icon_name = value
         elif pspec.name == "emblem-name":
             self._emblem_name = value
+        elif pspec.name == "secondary-emblem-name":
+            self._secondary_emblem_name = value
         elif pspec.name == "icon-tint":
             self._icon_tint = value
-            if self._icon_tint:
-                self._tint_color = Gdk.RGBA()
-                self._tint_color.parse(value)
-            else:
-                self._tint_color = None
         else:
             raise AttributeError("unknown property %s" % pspec.name)
 
@@ -69,6 +70,8 @@ class EmblemCellRenderer(Gtk.CellRenderer):
             return self._icon_name
         elif pspec.name == "emblem-name":
             return self._emblem_name
+        elif pspec.name == "secondary-emblem-name":
+            return self._secondary_emblem_name
         elif pspec.name == "icon-tint":
             return self._icon_tint
         else:
@@ -99,8 +102,8 @@ class EmblemCellRenderer(Gtk.CellRenderer):
                               pixbuf.get_width(), pixbuf.get_height())
             context.fill()
 
-            if self._tint_color:
-                c = self._tint_color
+            if self._icon_tint:
+                c = self._icon_tint
                 r, g, b = c.red, c.green, c.blue
                 # Figure out the difference between our tint colour and an
                 # empirically determined (i.e., guessed) satisfying luma and
@@ -121,6 +124,18 @@ class EmblemCellRenderer(Gtk.CellRenderer):
                                   cell_area.width, self._emblem_size)
                 context.fill()
 
+            if self._secondary_emblem_name:
+                pixbuf = self._get_pixbuf(
+                    self._secondary_emblem_name, self._emblem_size)
+                x_offset = self._icon_size - self._emblem_size
+                y_offset = self._icon_size - self._emblem_size + height_offset
+                context.set_operator(cairo.OPERATOR_OVER)
+                Gdk.cairo_set_source_pixbuf(
+                    context, pixbuf, x_offset, y_offset)
+                context.rectangle(
+                    x_offset, y_offset, cell_area.width, self._emblem_size)
+                context.fill()
+
         context.pop_group_to_source()
         context.set_operator(cairo.OPERATOR_OVER)
         context.paint()
