@@ -1,5 +1,3 @@
-# coding=UTF-8
-
 # Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
 # Copyright (C) 2010-2015 Kai Willadsen <kai.willadsen@gmail.com>
 #
@@ -19,30 +17,29 @@
 import atexit
 import functools
 import logging
-import tempfile
-import shutil
 import os
+import shutil
 import stat
 import sys
+import tempfile
 
 from gi.repository import Gdk
-from gi.repository import GLib
 from gi.repository import Gio
+from gi.repository import GLib
 from gi.repository import GObject
 from gi.repository import Gtk
 from gi.repository import Pango
 
-from meld import melddoc
-from meld import misc
-from meld import recent
 from meld import tree
-from meld import vc
-from meld.ui import gnomeglade
-from meld.ui import vcdialogs
-
 from meld.conf import _
-from meld.settings import settings, bind_settings
-from meld.vc import _null
+from meld.iohelpers import trash_or_confirm
+from meld.melddoc import MeldDoc
+from meld.misc import error_dialog, read_pipe_iter
+from meld.recent import RecentType
+from meld.settings import bind_settings, settings
+from meld.ui.gnomeglade import Component, ui_file
+from meld.ui.vcdialogs import CommitDialog, PushDialog
+from meld.vc import _null, get_vcs
 from meld.vc._vc import Entry
 
 log = logging.getLogger(__name__)
@@ -60,7 +57,7 @@ def cleanup_temp():
             if os.name == "nt":
                 os.chmod(f, stat.S_IWRITE)
             os.remove(f)
-        except:
+        except Exception:
             except_str = "{0[0]}: \"{0[1]}\"".format(sys.exc_info())
             print("File \"{0}\" not removed due to".format(f), except_str,
                   file=sys.stderr)
@@ -69,16 +66,17 @@ def cleanup_temp():
             assert (os.path.exists(f) and os.path.isabs(f) and
                     os.path.dirname(f) == temp_location)
             shutil.rmtree(f, ignore_errors=1)
-        except:
+        except Exception:
             except_str = "{0[0]}: \"{0[1]}\"".format(sys.exc_info())
             print("Directory \"{0}\" not removed due to".format(f), except_str,
                   file=sys.stderr)
 
+
 _temp_dirs, _temp_files = [], []
 atexit.register(cleanup_temp)
 
 
-class ConsoleStream(object):
+class ConsoleStream:
 
     def __init__(self, textview):
         self.textview = textview
@@ -115,13 +113,13 @@ COL_LOCATION, COL_STATUS, COL_OPTIONS, COL_END = \
 
 class VcTreeStore(tree.DiffTreeStore):
     def __init__(self):
-        tree.DiffTreeStore.__init__(self, 1, [str] * 5)
+        super().__init__(1, [str] * 5)
 
     def get_file_path(self, it):
         return self.get_value(it, self.column_index(tree.COL_PATH, 0))
 
 
-class VcView(melddoc.MeldDoc, gnomeglade.Component):
+class VcView(MeldDoc, Component):
 
     __gtype_name__ = "VcView"
 
@@ -131,13 +129,13 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         ('vc-merge-file-order', 'merge-file-order'),
     )
 
-    status_filters = GObject.property(
+    status_filters = GObject.Property(
         type=GObject.TYPE_STRV,
         nick="File status filters",
         blurb="Files with these statuses will be shown by the comparison.",
     )
-    left_is_local = GObject.property(type=bool, default=False)
-    merge_file_order = GObject.property(type=str, default="local-merge-remote")
+    left_is_local = GObject.Property(type=bool, default=False)
+    merge_file_order = GObject.Property(type=str, default="local-merge-remote")
 
     # Map for inter-tab command() calls
     command_map = {
@@ -153,12 +151,12 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     }
 
     def __init__(self):
-        melddoc.MeldDoc.__init__(self)
-        gnomeglade.Component.__init__(self, "vcview.ui", "vcview",
-                                      ["VcviewActions", 'liststore_vcs'])
+        MeldDoc.__init__(self)
+        Component.__init__(
+            self, "vcview.ui", "vcview", ["VcviewActions", 'liststore_vcs'])
         bind_settings(self)
 
-        self.ui_file = gnomeglade.ui_file("vcview-ui.xml")
+        self.ui_file = ui_file("vcview-ui.xml")
         self.actiongroup = self.VcviewActions
         self.actiongroup.set_translation_domain("meld")
         self.model = VcTreeStore()
@@ -167,8 +165,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self.treeview.set_model(self.model)
         self.treeview.get_selection().connect(
             "changed", self.on_treeview_selection_changed)
-        self.treeview.set_search_equal_func(
-            self.model.treeview_search_cb, None)
+        self.treeview.set_search_equal_func(tree.treeview_search_cb, None)
         self.current_path, self.prev_path, self.next_path = None, None, None
 
         self.name_column.set_attributes(
@@ -216,13 +213,13 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             pass
 
     def on_container_switch_in_event(self, ui):
-        super(VcView, self).on_container_switch_in_event(ui)
+        super().on_container_switch_in_event(ui)
         self._set_external_action_sensitivity(True)
         self.scheduler.add_task(self.on_treeview_cursor_changed)
 
     def on_container_switch_out_event(self, ui):
         self._set_external_action_sensitivity(False)
-        super(VcView, self).on_container_switch_out_event(ui)
+        super().on_container_switch_out_event(ui)
 
     def populate_vcs_for_location(self, location):
         """Display VC plugin(s) that can handle the location"""
@@ -241,7 +238,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             location = parent_location
         else:
             # existing parent directory was found
-            for avc in vc.get_vcs(location):
+            for avc in get_vcs(location):
                 err_str = ''
                 vc_details = {'name': avc.NAME, 'cmd': avc.CMD}
 
@@ -311,6 +308,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             return
 
         root = self.model.get_iter_first()
+        root_path = self.model.get_path(root)
 
         try:
             self.model.set_value(
@@ -319,12 +317,13 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             pass
 
         self.scheduler.add_task(self.vc.refresh_vc_state)
-        self.scheduler.add_task(self._search_recursively_iter(root))
+        self.scheduler.add_task(self._search_recursively_iter(root_path))
         self.scheduler.add_task(self.on_treeview_selection_changed)
         self.scheduler.add_task(self.on_treeview_cursor_changed)
 
     def get_comparison(self):
-        return recent.TYPE_VC, [self.location]
+        uris = [Gio.File.new_for_path(self.location)]
+        return RecentType.VersionControl, uris
 
     def recompute_label(self):
         self.label_text = os.path.basename(self.location)
@@ -332,7 +331,23 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         self.tooltip_text = _("%s: %s") % (_("Location"), self.location)
         self.label_changed()
 
-    def _search_recursively_iter(self, iterstart):
+    def _search_recursively_iter(self, start_path, replace=False):
+
+        # Initial yield so when we add this to our tasks, we don't
+        # create iterators that may be invalidated.
+        yield _("Scanning repository")
+
+        if replace:
+            # Replace the row at start_path with a new, empty row ready
+            # to be filled.
+            old_iter = self.model.get_iter(start_path)
+            file_path = self.model.get_file_path(old_iter)
+            new_iter = self.model.insert_after(None, old_iter)
+            self.model.set_value(new_iter, tree.COL_PATH, file_path)
+            self.model.set_path_state(new_iter, 0, tree.STATE_NORMAL, True)
+            self.model.remove(old_iter)
+
+        iterstart = self.model.get_iter(start_path)
         rootname = self.model.get_file_path(iterstart)
         display_prefix = len(rootname) + 1
         symlinks_followed = set()
@@ -419,7 +434,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def run_diff(self, path):
         if os.path.isdir(path):
-            self.emit("create-diff", [path], {})
+            self.emit("create-diff", [Gio.File.new_for_path(path)], {})
             return
 
         basename = os.path.basename(path)
@@ -432,8 +447,8 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         vc_entry = self.vc.get_entry(path)
         if vc_entry and vc_entry.state == tree.STATE_CONFLICT and \
                 hasattr(self.vc, 'get_path_for_conflict'):
-            local_label = _(u"%s — local") % basename
-            remote_label = _(u"%s — remote") % basename
+            local_label = _("%s — local") % basename
+            remote_label = _("%s — remote") % basename
 
             # We create new temp files for other, base and this, and
             # then set the output to the current file.
@@ -441,33 +456,33 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 conflicts = (tree.CONFLICT_THIS, tree.CONFLICT_MERGED,
                              tree.CONFLICT_OTHER)
                 meta['labels'] = (local_label, None, remote_label)
-                meta['tablabel'] = _(u"%s (local, merge, remote)") % basename
+                meta['tablabel'] = _("%s (local, merge, remote)") % basename
             else:
                 conflicts = (tree.CONFLICT_OTHER, tree.CONFLICT_MERGED,
                              tree.CONFLICT_THIS)
                 meta['labels'] = (remote_label, None, local_label)
-                meta['tablabel'] = _(u"%s (remote, merge, local)") % basename
+                meta['tablabel'] = _("%s (remote, merge, local)") % basename
             diffs = [self.vc.get_path_for_conflict(path, conflict=c)
                      for c in conflicts]
             temps = [p for p, is_temp in diffs if is_temp]
             diffs = [p for p, is_temp in diffs]
             kwargs = {
                 'auto_merge': False,
-                'merge_output': path,
+                'merge_output': Gio.File.new_for_path(path),
             }
             meta['prompt_resolve'] = True
         else:
-            remote_label = _(u"%s — repository") % basename
+            remote_label = _("%s — repository") % basename
             comp_path = self.vc.get_path_for_repo_file(path)
             temps = [comp_path]
             if self.props.left_is_local:
                 diffs = [path, comp_path]
                 meta['labels'] = (None, remote_label)
-                meta['tablabel'] = _(u"%s (working, repository)") % basename
+                meta['tablabel'] = _("%s (working, repository)") % basename
             else:
                 diffs = [comp_path, path]
                 meta['labels'] = (remote_label, None)
-                meta['tablabel'] = _(u"%s (repository, working)") % basename
+                meta['tablabel'] = _("%s (repository, working)") % basename
             kwargs = {}
         kwargs['meta'] = meta
 
@@ -475,7 +490,8 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             os.chmod(temp_file, 0o444)
             _temp_files.append(temp_file)
 
-        self.emit("create-diff", diffs, kwargs)
+        self.emit("create-diff",
+                  [Gio.File.new_for_path(d) for d in diffs], kwargs)
 
     def do_popup_treeview_menu(self, widget, event):
         if event:
@@ -509,9 +525,10 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         return False
 
     def on_filter_state_toggled(self, button):
-        active_action = lambda a: self.actiongroup.get_action(a).get_active()
         active_filters = [
-            k for k, v in self.state_actions.items() if active_action(v[0])]
+            k for k, (action_name, fn) in self.state_actions.items()
+            if self.actiongroup.get_action(action_name).get_active()
+        ]
 
         if set(active_filters) == set(self.props.status_filters):
             return
@@ -564,7 +581,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         files = [os.path.relpath(f, working_dir) for f in files]
         msg = shelljoin(command + files) + " (in %s)\n" % working_dir
         self.consolestream.command(msg)
-        readiter = misc.read_pipe_iter(
+        readiter = read_pipe_iter(
             command + files, workdir=working_dir,
             errorstream=self.consolestream)
         try:
@@ -573,7 +590,7 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
                 yield 1
                 result = next(readiter)
         except IOError as err:
-            misc.error_dialog(
+            error_dialog(
                 "Error running command",
                 "While running '%s'\nError: %s" % (msg, err))
             result = (1, "")
@@ -592,7 +609,20 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
         vc_command = self.command_map.get(command)
         return vc_command and hasattr(self.vc, vc_command)
 
-    def command(self, command, files):
+    def command(self, command, files, sync=False):
+        """
+        Run a command against this view's version control subsystem
+
+        This is the intended way for things outside of the VCView to
+        call in to version control methods, e.g., to mark a conflict as
+        resolved from a file comparison.
+
+        :param command: The version control command to run, taken from
+            keys in `VCView.command_map`.
+        :param files: File parameters to the command as paths
+        :param sync: If True, the command will be executed immediately
+            (as opposed to being run by the idle scheduler).
+        """
         if not self.has_command(command):
             log.error("Couldn't understand command %s", command)
             return
@@ -601,24 +631,30 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
             log.error("Invalid files argument to '%s': %r", command, files)
             return
 
+        runner = self.runner if not sync else self.sync_runner
         command = getattr(self.vc, self.command_map[command])
-        command(self.runner, files)
+        command(runner, files)
 
     def runner(self, command, files, refresh, working_dir):
         """Schedule a version control command to run as an idle task"""
         self.scheduler.add_task(
             self._command_iter(command, files, refresh, working_dir))
 
+    def sync_runner(self, command, files, refresh, working_dir):
+        """Run a version control command immediately"""
+        for it in self._command_iter(command, files, refresh, working_dir):
+            pass
+
     def on_button_update_clicked(self, obj):
         self.vc.update(self.runner)
 
     def on_button_push_clicked(self, obj):
-        response = vcdialogs.PushDialog(self).run()
+        response = PushDialog(self).run()
         if response == Gtk.ResponseType.OK:
             self.vc.push(self.runner)
 
     def on_button_commit_clicked(self, obj):
-        response, commit_msg = vcdialogs.CommitDialog(self).run()
+        response, commit_msg = CommitDialog(self).run()
         if response == Gtk.ResponseType.OK:
             self.vc.commit(
                 self.runner, self._get_selected_files(), commit_msg)
@@ -658,11 +694,17 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def on_button_delete_clicked(self, obj):
         files = self._get_selected_files()
         for name in files:
+            gfile = Gio.File.new_for_path(name)
+
             try:
-                gfile = Gio.File.new_for_path(name)
-                gfile.trash(None)
-            except GLib.GError as e:
-                misc.error_dialog(_("Error removing %s") % name, str(e))
+                trash_or_confirm(gfile)
+            except Exception as e:
+                error_dialog(
+                    _("Error deleting {}").format(
+                        GLib.markup_escape_text(gfile.get_parse_name()),
+                    ),
+                    str(e),
+                )
 
         workdir = os.path.dirname(os.path.commonprefix(files))
         self.refresh_partial(workdir)
@@ -684,17 +726,16 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
     def refresh_partial(self, where):
         if not self.actiongroup.get_action("VcFlatten").get_active():
             it = self.find_iter_by_name(where)
-            if it:
-                newiter = self.model.insert_after(None, it)
-                self.model.set_value(
-                    newiter, tree.COL_PATH, where)
-                self.model.set_path_state(newiter, 0, tree.STATE_NORMAL, True)
-                self.model.remove(it)
-                self.treeview.grab_focus()
-                self.treeview.get_selection().select_iter(newiter)
-                self.scheduler.add_task(self._search_recursively_iter(newiter))
-                self.scheduler.add_task(self.on_treeview_selection_changed)
-                self.scheduler.add_task(self.on_treeview_cursor_changed)
+            if not it:
+                return
+            path = self.model.get_path(it)
+
+            self.treeview.grab_focus()
+            self.vc.refresh_vc_state(where)
+            self.scheduler.add_task(
+                self._search_recursively_iter(path, replace=True))
+            self.scheduler.add_task(self.on_treeview_selection_changed)
+            self.scheduler.add_task(self.on_treeview_cursor_changed)
         else:
             # XXX fixme
             self.refresh()
@@ -741,9 +782,9 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_consoleview_populate_popup(self, textview, menu):
         buf = textview.get_buffer()
-        clear_cb = lambda *args: buf.delete(*buf.get_bounds())
         clear_action = Gtk.MenuItem.new_with_label(_("Clear"))
-        clear_action.connect("activate", clear_cb)
+        clear_action.connect(
+            "activate", lambda *args: buf.delete(*buf.get_bounds()))
         menu.insert(clear_action, 0)
         menu.insert(Gtk.SeparatorMenuItem(), 1)
         menu.show_all()
@@ -801,3 +842,10 @@ class VcView(melddoc.MeldDoc, gnomeglade.Component):
 
     def on_find_activate(self, *extra):
         self.treeview.emit("start-interactive-search")
+
+    def auto_compare(self):
+        modified_states = (tree.STATE_MODIFIED, tree.STATE_CONFLICT)
+        for it in self.model.state_rows(modified_states):
+            row_paths = self.model.value_paths(it)
+            paths = [p for p in row_paths if os.path.exists(p)]
+            self.run_diff(paths[0])
