@@ -31,6 +31,9 @@ def get_non_python_libs():
         # dll imported by dll dependencies expected to be auto-resolved later
         inst_root = [os.path.join(local_bin, 'libgtksourceview-3.0-1.dll')]
 
+        # required for communicating multiple instances
+        inst_lib.append(os.path.join(local_bin, 'gdbus.exe'))
+
         # gspawn-helper is needed for Gtk.show_uri function
         if platform.architecture()[0] == '32bit':
             inst_lib.append(os.path.join(local_bin, 'gspawn-win32-helper.exe'))
@@ -66,10 +69,13 @@ for data_dir in gtk_data_dirs:
             [os.path.join(local_data_subdir, file) for file in files]
         ))
 
-# add libgdk_pixbuf-2.0-0.dll manually to forbid auto-pulling of gdiplus.dll
 manually_added_libs = {
+    # add libgdk_pixbuf-2.0-0.dll manually to forbid auto-pulling of gdiplus.dll
     "libgdk_pixbuf-2.0-0.dll": os.path.join(sys.prefix, 'bin'),
-    }
+    # libcroco and librsvg are needed for SVG loading in gdkpixbuf
+    "libcroco-0.6-3.dll": os.path.join(sys.prefix, 'bin'),
+    "librsvg-2-2.dll": os.path.join(sys.prefix, 'bin'),
+}
 
 for lib, possible_path in manually_added_libs.items():
     local_lib = os.path.join(possible_path, lib)
